"""
For the irrep of U(d) with the given weight, viewed as a subspace of
(C^d)^{\otimes m}, this calculates the dimension of the subspace of the
weight zero subspace that is fixed by the action of S_d on tensor indices.

---

The calculations follow Theorem 2 of this paper:

Gay, David A. "Characters of the Weyl group of SU(n) on zero weight spaces and
centralizers of permutation representations." The Rocky Mountain Journal of
Mathematics 6.3 (1976): 449-455.
"""

from sage.all import *

import sym_character
import settings


def act_on_indices(g, l):
    return [l[g(i) - 1] for i in range(1, len(l) + 1)]


def irrep_dimension(t):
    if settings.SAFE:
        assert sum(t) == 0

    d = len(t)
    s = -t[-1]
    m = d * s

    Sd = SymmetricGroup(d)
    Sm = SymmetricGroup(m)

    # Build the S_m character corresponding to our irrep
    lam = tuple([x - t[-1] for x in t[:-1] if x != t[-1]])
    values = []
    for g in Sm.conjugacy_classes_representatives():
        mu = tuple(sorted(g.cycle_type(), reverse=True))
        values.append(sym_character.get_chi(lam, [], mu))
    chi = Sm.character(values)

    # Build the S_m character whose coefficient we will want to extract to
    # compute the dimension, by lifting from H to N(H) and inducing to S_m
    # (following the notation of the paper cited above)
    H_gens = []
    for i in range(d):
        for j in range(s - 1):
            H_gens.append((i * s + j + 1, i * s + j + 2))
    H = PermutationGroup(gens=H_gens)
    NH = Sm.normalizer(H)

    values = []
    x = sum([[i] * s for i in range(1, d + 1)], [])
    for g in NH.conjugacy_classes_representatives():
        g_x = act_on_indices(g, x)
        perm = []
        for i in range(d):
            perm.append(g_x[i * s])
            for j in range(s - 1):
                if settings.SAFE:
                    assert g_x[i * s + j] == g_x[i * s + j + 1]
        values.append(Permutation(perm).sign() ** s)
    mu_hat = NH.character(values)
    if settings.SAFE and s % 2 == 0:
        assert mu_hat == NH.trivial_character()

    mu_hat_Sm = mu_hat.induct(Sm)

    return chi.scalar_product(mu_hat_Sm)
