"""
Basic utilities for working with permutations.

These are often faster than converting to Sage objects and performing the
corresponding computation using its built-in functions.

The representation of a permutation here is as a list of length k whose entries
are 0, 1, ..., k - 1, where the list l represents the permutation pi(i) = l[i].
"""

import numpy as np


def to_matrix(p):
    """Returns the permutation matrix of p."""
    assert sorted(p) == range(max(p) + 1)

    d = len(p)
    S = np.zeros((d, d))

    for i, p_i in enumerate(p):
        S[i, p_i] = 1

    return S


def compose(p, q):
    """Composes two permutations."""
    return [q[x] for x in p]


def inverse(pi):
    """Returns the inverse of p."""
    indices = dict()
    for i, pi_i in enumerate(pi):
        indices[pi_i] = i
    return [indices[i] for i in range(len(pi))]


def cycle_type(p):
    """Returns a list of cycle lengths of p."""
    d = len(p)
    seen = set()
    ret = []

    while len(seen) < d:
        i = min([j for j in range(d) if j not in seen])
        seen.add(i)
        current_len = 1
        while p[i] not in seen:
            seen.add(p[i])
            i = p[i]
            current_len += 1

        ret.append(current_len)

    return ret


def sign(p):
    """Returns the sign of p."""
    d = len(p)
    seen = set()
    ret = 1

    while len(seen) < d:
        i = min([j for j in range(d) if j not in seen])
        seen.add(i)
        while p[i] not in seen:
            seen.add(p[i])
            i = p[i]
            ret *= -1

    return ret
