"""Utilities for working with S-matrices and the associated summation."""

import numpy as np
import scipy as sp
import scipy.sparse

from sage.all import *

import permutation
import settings


def det_Fd(d):
    """Determinant of the d-by-d Fourier matrix."""
    q_4 = d // 4
    r_4 = d % 4
    if r_4 == 0:
        ret = (-1) ** (q_4) * I
    elif r_4 == 1:
        ret = (-1) ** (q_4)
    elif r_4 == 2:
        ret = (-1) ** (q_4 + 1)
    elif r_4 == 3:
        ret = (-1) ** (q_4 + 1) * I
    else:
        assert False, (d, q_4, r_4)

    return ret


def evaluate_sum(S):
    """
    Evaluates the summation associated to S.

    Note that here we *include* the quantity det_Fd(d) ** (-s), for d the
    dimension and s the row/column sum of S, that appears when these summations
    come from irreps of U(d). We do *not* include the normalization of
    1 / factorial(d) ** 2.
    """
    d = S.nrows()
    if settings.SAFE:
        assert d == S.ncols()

    s = (S * vector([1] * d))[0]

    F = CyclotomicField(d)
    omega = F.gen()
    Sd = SymmetricGroup(d)

    x = range(d)
    vecs = []
    for p in Sd:
        vecs.append((p.sign() if s % 2 == 1 else 1, vector(p(x))))

    value = 0
    for sign1, v1 in vecs:
        for sign2, v2 in vecs:
            value += sign1 * sign2 * omega ** (v1 * S * v2)

    return value * det_Fd(d) ** (-s)


def cycle_to_start_with_min(c):
    """Cycles the entries of c to start with the smallest one."""
    if len(c) == 0:
        return c

    ix = 0
    smallest = c[0]
    for i, val in enumerate(c):
        if val < smallest:
            smallest = val
            ix = i

    return tuple(
        [c[ix + j] for j in range(len(c) - ix)] + [c[j] for j in range(ix)])


def canonical_matrix_of_key(k):
    """
    Converts a key to a canonical S-matrix (for testing).

    Returns an S-matrix that can come from a pair of partitions having the
    given key, as computed by key_and_sign_of_partitions below.
    """
    if type(k[0]) in (int, Integer):
        return identity_matrix(k[0])

    d = sum(k[0])
    s = len(k) + 1

    S = identity_matrix(d)

    c_cycles = []
    x = 1
    for c in k[0]:
        c_cycles.append(tuple(range(x, x + c)))
        x += c
    S += Permutation(c_cycles).inverse().to_matrix()

    for perm_cycles in k[1:]:
        seen = set()
        for c in perm_cycles:
            for i in c:
                seen.add(i)
        perm_cycles_with_fixed = list(perm_cycles)
        for i in range(1, d + 1):
            if i not in seen:
                perm_cycles_with_fixed.append((i,))
        S += Permutation(perm_cycles_with_fixed).inverse().to_matrix()

    return S


def matrix_of_partitions(p1, p2):
    """Returns the S-matrix of a pair of partitions."""
    d = Integer(len(set(p1)))
    S = zero_matrix(d, d)

    for i, j in zip(p1, p2):
        S[i, j] += 1

    return S


def key_and_sign_of_partitions(p1, p2):
    """Returns a key determining the S-matrix summation up to sign."""
    d = Integer(len(set(p1)))
    s = Integer(len(p1)) / d

    if settings.SAFE:
        assert s in (1, 2, 3)

    if s == 1:
        perm1 = [None] * d
        for i, j in zip(p1, p2):
            perm1[i] = j

        return ((d,), permutation.sign(perm1))
    elif s == 2:
        S = sp.sparse.coo_matrix(
            (np.ones(d * s), (p1, p2)),
            shape=(d, d)).tocsr()

        perm1 = sp.sparse.csgraph.maximum_bipartite_matching(
            S, perm_type='column')
        S[range(d), perm1] -= 1
        S.eliminate_zeros()

        perm2 = sp.sparse.csgraph.maximum_bipartite_matching(
            S, perm_type='column')

        return ((
            tuple(sorted(map(len, Permutation([
                x + 1
                for x in permutation.compose(
                    permutation.inverse(perm1), perm2)]
            ).cycle_tuples(singletons=True)), reverse=True)),
        ), 1)
    elif s == 3:
        S = sp.sparse.coo_matrix(
            (np.ones(d * s), (p1, p2)),
            shape=(d, d)).tocsr()

        perm1 = sp.sparse.csgraph.maximum_bipartite_matching(
            S, perm_type='column')
        S[range(d), perm1] -= 1
        S.eliminate_zeros()

        perm2 = sp.sparse.csgraph.maximum_bipartite_matching(
            S, perm_type='column')
        S[range(d), perm2] -= 1
        S.eliminate_zeros()

        perm3 = sp.sparse.csgraph.maximum_bipartite_matching(
            S, perm_type='column')

        perm1_inv = permutation.inverse(perm1)

        g = Permutation([
            x + 1
            for x in permutation.compose(perm1_inv, perm2)])
        g_cycles = sorted(
            g.cycle_tuples(singletons=True),
            key=lambda l: len(l), reverse=True)
        renaming = range(d)
        ix = 1
        for c in g_cycles:
            for i in c:
                renaming[i - 1] = ix
                ix += 1

        pi3_cycles = Permutation([
            i + 1 for i in permutation.compose(
                perm1_inv, perm3)]).cycle_tuples(singletons=False)
        pi3_cycles_renamed = tuple(sorted([
            cycle_to_start_with_min(tuple([renaming[i - 1] for i in l]))
            for l in pi3_cycles]))

        return (
            (tuple(map(len, g_cycles)),
             pi3_cycles_renamed),
            permutation.sign(perm1))
