import random as rn
import unittest

from sage.all import *

import s_matrix


SEED = 0
NUM_RANDOM_CASES = 20
# Place to put bad cases from debugging to make sure they are covered:
FIXED_PARTITION_CASES = [
    ((0, 0, 0, 1, 1, 2, 1, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5),
     (0, 0, 0, 1, 1, 2, 2, 1, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5))]


def random_balanced_partition(d, s):
    p = sum([[i] * s for i in range(d)], [])
    rn.shuffle(p)
    return p


def generate_partition_cases(d, s):
    rn.seed(SEED)

    for p1, p2 in FIXED_PARTITION_CASES:
        if len(set(p1)) == d and len(p1) / len(set(p1)) == s:
            yield p1, p2

    for _ in range(NUM_RANDOM_CASES):
        yield random_balanced_partition(d, s), random_balanced_partition(d, s)


class TestCanonicalMatrix(unittest.TestCase):
    def test_canonical_matrix_of_key_s1(self):
        return
        d = 6
        S = s_matrix.canonical_matrix_of_key((d,))
        self.assertEqual(
            S,
            matrix([
                [1, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0],
                [0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 1]]))

    def test_canonical_matrix_of_key_s2(self):
        return
        S = s_matrix.canonical_matrix_of_key( ((1,2,3),) )
        self.assertEqual(
            S,
            matrix([
                [2, 0, 0, 0, 0, 0],
                [0, 1, 1, 0, 0, 0],
                [0, 1, 1, 0, 0, 0],
                [0, 0, 0, 1, 0, 1],
                [0, 0, 0, 1, 1, 0],
                [0, 0, 0, 0, 1, 1]]))

        S = s_matrix.canonical_matrix_of_key( ((4,2,),) )
        self.assertEqual(
            S,
            matrix([
                [1, 0, 0, 1, 0, 0],
                [1, 1, 0, 0, 0, 0],
                [0, 1, 1, 0, 0, 0],
                [0, 0, 1, 1, 0, 0],
                [0, 0, 0, 0, 1, 1],
                [0, 0, 0, 0, 1, 1]]))

    def test_canonical_matrix_of_key_s3(self):
        return
        S = s_matrix.canonical_matrix_of_key(
            (
                (1,2,3),
                # Cycle type needs to be read right to left, i.e., here the
                # first cycle is 1 => 2 => 5 => 1
                ((1, 5, 2), (3, 4, 6))
            ))
        self.assertEqual(
            S,
            matrix([
                [2, 1, 0, 0, 0, 0],
                [0, 1, 1, 0, 1, 0],
                [0, 1, 1, 0, 0, 1],
                [0, 0, 1, 1, 0, 1],
                [1, 0, 0, 1, 1, 0],
                [0, 0, 0, 1, 1, 1]]))


class TestKeyConsistency(unittest.TestCase):
    def test_round_trip_permutation(self):
        """Tests that matrix => key => canonical matrix gives permutation"""
        rn.seed(0)

        for d in range(2, 6):
            for s in range(1, 4):
                Sd = SymmetricGroup(d)

                for ix, (p1, p2) in enumerate(generate_partition_cases(d, s)):
                    S = s_matrix.matrix_of_partitions(p1, p2)

                    key, sign = s_matrix.key_and_sign_of_partitions(p1, p2)
                    S_canon = s_matrix.canonical_matrix_of_key(key)

                    flag = False

                    for pi1 in Sd:
                        for pi2 in Sd:
                            if (Permutation(pi1).to_matrix() *
                                S *
                                Permutation(pi2).to_matrix() == S_canon):
                                flag = True
                                break
                        if flag:
                            break

                    self.assertTrue(
                        flag,
                        msg=(
                            str(p1) + ', ' + str(p2) + '\n\n' +
                            str(key) + '\n\n' +
                            str(S) + '\n\n' +
                            str(S_canon) + '\n'
                        ))

                    # It would seem to make sense to test that sign and
                    # (p1.sign() * p2.sign()) ** 2 coincide here, but that
                    # actually doesn't need to be the case necessarily, so
                    # long as the value of the sum is zero.

    def test_round_trip_value(self):
        """Testing matrix => key => canonical matrix equality"""
        return
        rn.seed(0)
        num_experiments = 20

        for d in range(2, 6):
            for s in range(1, 4):
                for p1, p2 in generate_partition_cases(d, s):
                    S = s_matrix.matrix_of_partitions(p1, p2)
                    key, sign = s_matrix.key_and_sign_of_partitions(p1, p2)

                    S_canon = s_matrix.canonical_matrix_of_key(key)

                    sum_direct = s_matrix.evaluate_sum(S)
                    sum_canon = s_matrix.evaluate_sum(S_canon) * sign

                    self.assertEqual(
                        sum_direct,
                        sum_canon,
                        msg=(
                            str(p1) + ', ' + str(p2) + '\n\n' +
                            str(S) + '\n\n' +
                            str(S_canon) + '\n' + str(sign) + '\n\n' +
                            str(sum_direct) + ' != ' + str(sum_canon)
                        ))


if __name__ == '__main__':
    unittest.main()
