"""
This script should output True if and only if script_irrep_dimension.py outputs
0 with the same input. It works more explicitly by checking if the projection
of every basis element of the weight zero subspace to the subspace invariant
under the action of the symmetric group is zero. This can require quite a lot
of memory and gets slow when the last term is -3.

Usage: sage -python ./script_check_basis_zero.py 1 1 0 0 -1 -1
"""

import argparse
import collections as co

import numpy as np
from sage.all import *

import partition
import settings
import tableau


PARSER = argparse.ArgumentParser(
    description=('Check whether $\dim(\widetilde{V}_{\pi}) = 0$ for an ' +
                 'irrep $\pi$ by explicitly computing projections of basis ' +
                 'elements.'))
PARSER.add_argument(
    'irrep_weights',
    nargs='*',
    type=int,
    help='integer weight sequence of the irrep to check')
PARSER.add_argument(
    '--explicit_tensor',
    action='store_true',
    help='if passed, use a numpy tensor instead of a sparse representation')
ARGS = PARSER.parse_args()


def basis_tensor(n, ixs):
    ret = np.zeros([n] * len(ixs), dtype=np.int32)
    ret[tuple(ixs)] = 1
    return ret


def main():
    # Retrieve and process arguments
    t = ARGS.irrep_weights

    if settings.SAFE:
        assert sum(t) == 0

    d = len(t)
    s = -t[-1]
    m = d * s

    lam_min = t[-1]
    lam = tuple([x - t[-1] for x in t])
    mu = tuple([-t[-1] for x in t])

    Sd = SymmetricGroup(d)

    # Get canonicalized semistandard tableaux
    all_ssyt = SemistandardTableaux(lam, mu).list()
    if settings.VERBOSITY['sizes']:
        print 'Total SSYTs:', len(all_ssyt)
    canon_ssyt = []
    canon_partitions = []
    for T in all_ssyt:
        this_canonical_partition = partition.canonical_partition(T.entries())
        if this_canonical_partition not in canon_partitions:
            canon_ssyt.append(T)
            canon_partitions.append(this_canonical_partition)
    if settings.VERBOSITY['sizes']:
        print 'Canonical SSYTs:', len(canon_ssyt)

    # Build row and column stabilizer
    T_std = tableau.default_standard_tableau(lam)
    R = T_std.row_stabilizer()
    C = T_std.column_stabilizer()
    if settings.VERBOSITY['sizes']:
        print 'Row stabilizer size:', len(R)
        print 'Column stabilizer size:', len(C)
        print 'Number of Young symmetrizer terms:', len(R) * len(C)

    # Build list of Young symmetrizer terms
    symmetrizer_terms = []
    for r in R:
        for c in C:
            symmetrizer_terms.append((c.sign(), c(r(range(1, m + 1)))))

    # Check basis elements
    for T in canon_ssyt:
        if settings.VERBOSITY['progress']:
            print 'Checking', T
        entries = T.entries()

        if ARGS.explicit_tensor:
            v = np.zeros([d] * m, dtype=np.int32)
            for i, (sign, perm) in enumerate(symmetrizer_terms):
                for perm2 in SymmetricGroup(d):
                    v += (sign * basis_tensor(d, [
                        perm2(entries[perm[i] - 1]) - 1
                        for i in range(m)]) * perm2.sign() ** s)

            if np.sum(v * v) > 0:
                print False
                break
        else:
            coeff_dict = co.defaultdict(int)

            for sign, perm in symmetrizer_terms:
                canonical, achieving_sign = partition.canonical_partition(
                    [entries[i - 1] for i in perm], include_sign=True)
                coeff_dict[canonical] += (
                    sign * achieving_sign ** s)

            if any(x != 0 for x in coeff_dict.itervalues()):
                print False
                break
    else:
        print True


if __name__ == '__main__':
    main()
