"""
Main script to check the positivity condition for a given irrep.

Usage:

   sage -python script_check_positivity.py 3 0 0 0 0 -3 \
     --ssyt_basis_index=1 \
     --outer_slice=0:10 \
     --s_matrix_values_file=data/s_matrix_values_6_3.csv
"""

import argparse
import functools
import itertools
import time

import numpy as np
from sage.all import *

import partition
import s_matrix
import settings
import tableau


PARSER = argparse.ArgumentParser(
    description='Check the positivity condition for a given irrep.')
PARSER.add_argument(
    't',
    nargs='*',
    type=int,
    help='integer weight sequence of the irrep for which to check positivity')
PARSER.add_argument(
    '--ssyt_basis_index',
    type=int,
    required=True,
    help=('index in Sage\'s ordering of the semistandard Young tableau ' +
          'corresponding to the irrep basis element we use to build a test ' +
          'vector ($v$ in the paper\'s notation)'))
PARSER.add_argument(
    '--outer_slice',
    type=str,
    default=None,
    help=('if the computation is to be divided into parts at the outermost ' +
          'summation, then this should be of the form `a:b` for the slice ' +
          'of indices to include in the summation (as in ordinary Python ' +
          'syntax)'))
PARSER.add_argument(
    '--s_matrix_values_file',
    type=str,
    default=None,
    help=('file of precomputed S-matrix summations (from ' +
          'script_s_matrix.py), if the default location of the provided ' +
          'files is not to be used'))
ARGS = PARSER.parse_args()


def contains_repetition(l):
    """Returns whether l contains a repeated element."""
    return len(set(l)) < len(l)


# To avoid some annoying Sage casting issues:
def int_len(l):
    """List length as a Sage Integer."""
    return Integer(len(l))


def main():
    # Retrieve inputs
    t = tuple(map(Integer, ARGS.t))
    ssyt_basis_ix = Integer(ARGS.ssyt_basis_index)
    outer_slice_ixs = None
    if ARGS.outer_slice is not None:
        outer_slice_ixs = tuple(map(int, ARGS.outer_slice.split(':')))

    d = len(t)
    s = -t[-1]
    m = d * s

    lam_min = t[-1]
    lam = tuple([x - t[-1] for x in t])
    mu = tuple([-t[-1] for x in t])

    # Retrieve the SSYT indexing the basis element to work with and the
    # associated partition.
    T = SemistandardTableaux(lam, mu)[ssyt_basis_ix]
    p = partition.canonical_partition(T.entries())
    p_chunks = []
    for i in range(d):
        p_chunks.append([j for j in range(d * s) if p[j] == i])

    if settings.VERBOSITY['info']:
        print 'Using SSYT', str(T), 'with canonical partition', str(p)

    Sd = SymmetricGroup(d)
    Sm = SymmetricGroup(m)

    # Retrieve precomputed S-matrix summations
    if ARGS.s_matrix_values_file is None:
        s_matrix_values_file = './data/s_matrix_values_%d_%d.csv' % (d, s)
    else:
        s_matrix_values_file = ARGS.s_matrix_values_file

    s_matrix_values = dict()
    with open(s_matrix_values_file, 'r') as f:
        line = f.readline()
        while line:
            k, v = sage_eval(line.strip())
            # Add in the normalization here that is skipped in the S matrix
            # computations
            s_matrix_values[k] = Integer(v) / factorial(d) ** 2
            line = f.readline()

    # Computations of the row and column stabilizer to express the Young
    # symmetrizer
    T_std = tableau.default_standard_tableau(lam)
    R = T_std.row_stabilizer()
    R_gap = libgap(R)
    C = T_std.column_stabilizer()
    if settings.VERBOSITY['sizes']:
        print ''
        print 'Row stabilizer size:', len(R)
        print 'Column stabilizer size:', len(C)
        print 'Number of Young symmetrizer terms:', len(R) * len(C)
    T_col_ixs = T_std.conjugate().to_list()

    # Compute coset representatives for the stabilizer of the SSYT rows in
    # the stabilizer of the underlying Young tableau rows. Contributions
    # of all such terms will have the same value, as detailed in Appendix D
    # of the paper
    T_row_stab_gap = libgap(tableau.ssyt_row_stabilizer(T))
    R_quotient = [
        coset.Representative().sage().inverse()
        for coset in R_gap.RightCosets(T_row_stab_gap)]
    R_quotient_inv = [
        coset.Representative().sage()
        for coset in R_gap.RightCosets(T_row_stab_gap)]
    if outer_slice_ixs is not None:
        R_quotient_inv = R_quotient_inv[outer_slice_ixs[0]:outer_slice_ixs[1]]
    if settings.VERBOSITY['sizes']:
        print 'SSYT row stabilizer reduces size to:', len(R_quotient)

    partition_pair_key_dict = dict()

    # We keep two values: $v^{\top} M v$ where M is the image of the Fourier
    # matrix under the irrep we are working with, and $\|v\|^2 = v^{\top} I v$
    M_value = 0
    Id_value = 0

    last_time = time.time()

    if settings.VERBOSITY['progress']:
        print ''

    C_with_signs = [(c, c.sign()) for c in C]

    # Again, see Appendix D for the details of why the summations may be
    # reorganized in this way
    for ix_r1, r1 in enumerate(R_quotient_inv):
        # In this case the column symmetrizer will kill all entries:
        if any(contains_repetition([r1(p)[i - 1] for i in l])
               for l in T_col_ixs):
            continue

        for ix_r2, r2 in list(enumerate(R_quotient)):
            # Likewise here:
            if any(contains_repetition([r2(p)[i - 1] for i in l])
                   for l in T_col_ixs):
                continue

            r2_p = r2(p)

            for ix_c, (c, c_sign) in enumerate(C_with_signs):
                if (settings.VERBOSITY['progress'] and
                    ix_c % settings.CHECK_POSITIVITY_LOG_EVERY == 0):
                    this_time = time.time()

                    log_line_parts = [
                        str(ix_r1 + 1) +  ' / ' + str(len(R_quotient)),
                        str(ix_r2 + 1) +  ' / ' + str(len(R_quotient)),
                        str(ix_c + 1) +  ' / ' +  str(len(C)),
                        str(len(partition_pair_key_dict))]
                    if settings.VERBOSITY['timing']:
                        log_line_parts.append(str(this_time - last_time) + 's')
                    print '; '.join(log_line_parts)

                    last_time = this_time

                p2_perm = r1(c(r2_p))
                p2_canon, p2_achieving_sign = partition.canonical_partition(
                    p2_perm, include_sign=True)

                # Fast sorting for small s
                p2_canon_sort = [None for _ in xrange(d * s)]
                for i in xrange(d):
                    if s == 1:
                        pass
                    if s == 2:
                        p2_canon_sort[p_chunks[i][0]] = min(
                            p2_canon[p_chunks[i][0]],
                            p2_canon[p_chunks[i][1]])
                        p2_canon_sort[p_chunks[i][1]] = max(
                            p2_canon[p_chunks[i][0]],
                            p2_canon[p_chunks[i][1]])
                    elif s == 3:
                        p2_canon_sort[p_chunks[i][0]] = min(
                            p2_canon[p_chunks[i][0]],
                            p2_canon[p_chunks[i][1]],
                            p2_canon[p_chunks[i][2]])
                        p2_canon_sort[p_chunks[i][2]] = max(
                            p2_canon[p_chunks[i][0]],
                            p2_canon[p_chunks[i][1]],
                            p2_canon[p_chunks[i][2]])
                        p2_canon_sort[p_chunks[i][1]] = (
                            p2_canon[p_chunks[i][0]] +
                            p2_canon[p_chunks[i][1]] +
                            p2_canon[p_chunks[i][2]] -
                            p2_canon_sort[p_chunks[i][2]] -
                            p2_canon_sort[p_chunks[i][0]])
                    else:
                        for j, val in itertools.izip(
                                p_chunks[i],
                                sorted([p2_canon[j] for j in p_chunks[i]])):
                            p2_canon_sort[j] = val
                p2_canon_sort = tuple(p2_canon_sort)

                # Retrieve S-matrix summation for this term
                s_value = partition_pair_key_dict.get(p2_canon_sort)
                if s_value is None:
                    s_key, s_sign = s_matrix.key_and_sign_of_partitions(
                        p, p2_canon_sort)
                    s_value = s_sign * s_matrix_values[s_key]
                    partition_pair_key_dict[p2_canon_sort] = s_value

                # Update the two running values
                M_value += (
                    s_value
                    * c_sign
                    * p2_achieving_sign ** s
                )

                if p == p2_canon:
                    Id_value += (
                        c.sign()
                        * p2_achieving_sign ** s
                    ) / factorial(d)

    # Normalization from $\sqrt{d}$ terms in the S-matrix summations for the
    # Fourier matrix
    M_value = M_value / (d ** (m / 2))

    final_value = M_value + Id_value / d

    if settings.VERBOSITY['progress']:
        print ''

    if outer_slice_ixs is None:
        if Id_value > 0 and final_value >= 0:
            print 'Success!'

        if Id_value > 0:
            print 'v^T I v = |v|^2 = %s > 0' % str(Id_value)
        else:
            print 'Error: basis vector projected to zero'

        if final_value >= 0:
            print 'v^T (M + I/d) v = %s >= 0' % str(final_value)
        else:
            print 'Error: v^T (M + I/d) v = %s < 0' % str(final_value)
    else:
        print M_value
        print Id_value
        print final_value


if __name__ == '__main__':
    main()
