"""
Computes the dimension of the subspace invariant under the $S_d$ action of the
weight zero subspace of an irrep.

Usage:
1. To compute the dimension for a single irrep:

   sage -python script_irrep_dimension.py single 2 0 0 0 0 -2

2. To compute the dimension for all irreps with a given number of weights whose
   absolute values add up to at most a given bound:

   sage -python script_irrep_dimension.py r_up_to 6 3
"""

import argparse

from sage.all import *

import irrep_dimension


PARSER = argparse.ArgumentParser(
    description='Compute $\dim(\widetilde{V}_{\pi})$ for an irrep $\pi$.')
PARSER.add_argument(
    'mode',
    type=str,
    help=('either `single` to compute the dimension for a single irrep, or ' +
          '`r_up_to` to compute the dimension for all irreps with the sum ' +
          'of absolute values of the weights at most some value'))
PARSER.add_argument(
    'params',
    nargs='*',
    type=int,
    help=('if `mode` is `single`, then the weights of the irrep to compute ' +
          'the dimension for; if `mode` is `r_up_to`, then the number of ' +
          'weights ($d$ in the paper\'s notation) and the bound on the sum ' +
          'of absolute values of weights'))
ARGS = PARSER.parse_args()

MODES = ('single', 'r_up_to')


def main():
    assert ARGS.mode in MODES

    if ARGS.mode == 'single':
        t = tuple(ARGS.params)
        print irrep_dimension.irrep_dimension(t)
    elif ARGS.mode == 'r_up_to':
        d = ARGS.params[0]
        max_r = ARGS.params[1]

        for r in range(1, max_r + 1):
            parts = Partitions(r)
            for p1 in parts:
                for p2 in parts:
                    if len(p1) + len(p2) <= d:
                        t = tuple(
                            p1
                            + [0] * (d - len(p1) - len(p2))
                            + [-i for i in p2[::-1]])
                        dim = irrep_dimension.irrep_dimension(t)

                        if dim > 0:
                            print t, dim


if __name__ == '__main__':
    main()
