"""
Precomputes S-matrix summations.

Usage:

   sage -python \
     ./script_s_matrix.py 6 3 \
     --output_file=s_matrix_values_6_3.csv
"""

import argparse
import collections as co
import itertools as it
import sys

from sage.all import *

import settings
import s_matrix


PARSER = argparse.ArgumentParser(description='Precompute S-matrix summations.')
PARSER.add_argument(
    'd',
    type=int,
    help='the dimension of the S matrices')
PARSER.add_argument(
    's',
    type=int,
    help='the sum of each row and column of the S matrices')
PARSER.add_argument(
    '--output_file',
    type=str,
    required=True,
    help='the file to write output to')
ARGS = PARSER.parse_args()


def main():
    d = ARGS.d
    s = ARGS.s

    if settings.SAFE:
        assert s in (2, 3)

    F = CyclotomicField(d)
    omega = F.gen()
    Sd = SymmetricGroup(d)

    keys = []
    values = []

    if s == 2:
        num_keys = len(Sd.conjugacy_classes())
    elif s == 3:
        num_keys = len(Sd.conjugacy_classes()) * len(Sd)

    if settings.VERBOSITY['sizes']:
        print 'Total keys:', num_keys

    x = range(d)
    vecs = []
    for p in Sd:
        vecs.append((p.sign() if s % 2 == 1 else 1, vector(p(x))))

    keys = []
    values = []
    ix = 1
    if s == 2:
        for conj_class in Sd.conjugacy_classes():
            if settings.VERBOSITY['progress']:
                print ix, '/', num_keys

            key = (tuple(conj_class.representative().cycle_type()),)
            keys.append(key)

            S = identity_matrix(d)

            c_cycles = []
            x = 1
            for c in key[0]:
                c_cycles.append(tuple(range(x, x + c)))
                x += c
            S += Permutation(c_cycles).to_matrix()

            values.append(s_matrix.evaluate_sum(S))

            ix += 1
    elif s == 3:
        for conj_class in Sd.conjugacy_classes():
            for perm in Sd:
                if settings.VERBOSITY['progress']:
                    print ix, '/', num_keys

                key = (
                    tuple(conj_class.representative().cycle_type()),
                    tuple(perm.cycle_tuples()))
                keys.append(key)

                S = identity_matrix(d)
                S += Permutation(perm).to_matrix()

                c_cycles = []
                x = 1
                for c in key[0]:
                    c_cycles.append(tuple(range(x, x + c)))
                    x += c
                S += Permutation(c_cycles).to_matrix()

                values.append(s_matrix.evaluate_sum(S))

                ix += 1

    with open(ARGS.output_file, 'w') as f:
        for key, value in zip(keys, values):
            f.write(str(key) + ',' + str(value) + '\n')


if __name__ == '__main__':
    main()
