"""
The functions get_chi and get_chi_recursive compute symmetric group characters
recursively using the Murnaghan-Nakayama rule. They are patched in from this
contribution to the Sage beta (at the time of writing):

git.sagemath.org/sage.git/commit?id=04d133ded6fb7ce9e523e371b335f003bf58d5da

The style of this code is left as-is from the patch.
"""

from sage.all import *


def get_chi(lda, mu, alpha):
    """
    Return `\chi^{\lambda/\mu}(\alpha)` value in Murnaghan-Nakayama formula
    (Enumerative Combinatorics (Stanley) equation 7.75)

    `\lambda`: list of non-increasing integers

    `\mu`: list of non-increasing integers, pass empty list to `\mu` if not
           applicable.

    `\alpha`: list of non-negative integers (weak composition)
    """
    if len(lda) >= len(mu):
        # padding mu with zero (same length with lda)
        mu += [0] * (len(lda)-len(mu))
    else:
        return None

    # ensure lda and mu are integer partition
    for i in range(1, len(lda)):
        if lda[i] > lda[i-1] or mu[i] > mu[i-1] or \
            lda[i] < 0 or mu[i] < 0:
            return None

    # ensure alpha is a weak composition
    for i in range(1, len(alpha)):
        if alpha[i] < 0:
            return None

    # alpha weak composition -> composition
    alpha = [x for x in alpha if x != 0]


    expectedBlocks = sum(alpha)

    # ensure each row interval exists
    for i in range(len(lda)):
        if mu[i] > lda[i]:
            return None
        else:
            expectedBlocks -= lda[i]-mu[i]

    # ensure sum(lda)-sum(mu)=sum(alpha)
    if expectedBlocks != 0:
        return None

    return get_chi_recursive(lda, mu, alpha, {"[][][]": 1})


def get_chi_recursive(lda, mu, alpha, refDict):
    """
    Return `\chi^{\lambda/\mu}(\alpha)` value in Murnaghan-Nakayama formula
    (Recursive subroutine for ``get_chi()``)

    `\lambda`: list of non-increasing integers

    `\mu`: list of non-increasing integers, pass empty list to `\mu` if not
           applicable.

    `\alpha`: list of positive integers (composition)

    refDict: dictionary with the format {
           str([``magn``][``conn``][``alpha``]):    ``ref_chi_value``}, where

    - ``magn`` is a list of number of available blocks each row

    - ``conn`` is a list of number of connected blocks between each pair of
      consecutive rows, ``conn[i] = 0`` when row `i-1` and `i` are not
      connected.

    - both ``magn`` and ``conn`` assumes `\lambda[i]` > `\mu[i]` for all `i`.

    - ``ref_chi_value`` is the value of `\chi` given ``magn``, ``conn`` and
      ``alpha``
    """
    # return None if not connected
    # if diminish case occur (normal check useless) and all later rows also
    # diminish, then good
    # if diminish case occur (normal check useless) and some later rows do
    # not diminish, then bad
    # if diminish case do not occur, then good/bad depends on normal check

    # remove all rows with no available blocks
    i = 0
    while i < len(lda):
        if lda[i] == mu[i]:
            del lda[i]
            del mu[i]
        else:
            i += 1

    # use the following for refDict
    # i.e. if magnitude/connection/alpha are the same -> same psi
    magnitude = []
    connection = []

    i = 0
    while i < len(lda):
        magnitude.append(lda[i]-mu[i])
        if i > 0:
            # diff between current lda and last mu
            diff = lda[i]-mu[i-1]
            if diff > 0:
                connection.append(diff)
            else:
                connection.append(0)
        i += 1

    refPsi = refDict.get(str(magnitude)+str(connection)+str(alpha), None)
    if refPsi is not None:
        return refPsi

    # at this point nothing is found in refPsi, which
    # also ensures non empty magnitude
    max_width_index = lda[0] - 1
    max_height_index = len(lda) - 1

    # start finding border (row, col) (index 0 based)
    border = []
    numBlocks = []

    currentPt = (0, max_width_index)
    # borderEnd = (max_height_index, mu[max_height_index])
    # extra pt at the end of border
    extraPt = (max_height_index, mu[max_height_index]-1)

    while currentPt != extraPt:
        # good pt
        if currentPt[1] >= mu[currentPt[0]]:
            border.append(currentPt)
        # bad pt (in mu)
        else:
            border.append(currentPt)
            numBlocks += list(range(len(border)-len(numBlocks)-1, -1, -1))

        # if can go down
        if ((currentPt[0] < max_height_index) and
                (currentPt[1] == lda[currentPt[0]+1]-1)):
            currentPt = (currentPt[0]+1, currentPt[1])
        # cannot go down, then go left
        else:
            currentPt = (currentPt[0], currentPt[1]-1)

    # append the extra bad block
    border.append(extraPt)
    numBlocks += list(range(len(border)-len(numBlocks)-1, -1, -1))

    # possible start point's index in border list
    # nothing to the right and is a good pt
    start = [0]
    for i in range(1, len(border)):
        if border[i][1]+1 == lda[border[i][0]] and numBlocks[i] > 0:
            start.append(i)

    # use those points to derive possible subTableaux
    # notice that the extra point ensures that ptIndex + numBlocks[ptIndex]
    # at any start pt always stays in border
    length = alpha[0]
    psi = 0
    for ptIndex in start:
        if length <= numBlocks[ptIndex]:
            if border[ptIndex+length][1] < border[ptIndex+length-1][1]:
                subLda = list(lda)
                subMu = list(mu)
                for i in range(ptIndex, ptIndex+length):
                    subLda[border[i][0]] -= 1

                # no need to check length of alpha, [1:] won't throw an error

                subPsi = get_chi_recursive(subLda, subMu, alpha[1:], refDict)
                stripHeight = border[ptIndex+length-1][0] - border[ptIndex][0]
                psi += subPsi * (-1)**stripHeight

    refDict[str(magnitude)+str(connection)+str(alpha)] = psi
    return psi
